#ifndef MAGIC_TYPE
  #ifdef __MQL5__
    #define MAGIC_TYPE long
  #else // __MQL5__
    #define MAGIC_TYPE int
  #endif // __MQL5__
#endif // MAGIC_TYPE

enum ENUM_ONTESTERCUSTOM
{
  ONTESTERCUSTOM_GAIN,                // Gain
  ONTESTERCUSTOM_MAXDD,               // MaxDD
  ONTESTERCUSTOM_RECOVERYFACTOR,      // Recovery Factor
  ONTESTERCUSTOM_GAIN_FIXDD,          // Gain with Fix DD
  ONTESTERCUSTOM_RISK_FIXDD,          // Risk with Fix DD
  ONTESTERCUSTOM_RECOVERYFACTOR_FIXDD // Recovery Factor with Fix DD
};

class ONTESTERCUSTOM
{
private:
  double Positions[];

#define MACROS_MT4                                                                   \
{                                                                                    \
  const int Size = ::ArrayResize(this.Positions, OrdersHistoryTotal());              \
                                                                                     \
  for (int i = 0; i < Size; i++)                                                     \
    if (OrderSelect(i, SELECT_BY_POS, MODE_HISTORY) && (OrderType() <= OP_SELL) &&   \
        ((Magic == -1) || (OrderMagicNumber() == Magic)) && (OrderSymbol() == Symb)) \
      this.Positions[Amount++] = OrderType() ? OrderOpenPrice() / OrderClosePrice()  \
                                             : OrderClosePrice() / OrderOpenPrice(); \
}

#define MACROS_MT5                                                                            \
{                                                                                             \
  if (::HistorySelect(0, INT_MAX))                                                            \
  {                                                                                           \
    const int Size = ::ArrayResize(this.Positions, ::HistoryDealsTotal());                    \
                                                                                              \
    for (int i = 0; i < Size; i++)                                                            \
    {                                                                                         \
      const ulong Ticket = ::HistoryDealGetTicket(i);                                         \
      const ENUM_DEAL_TYPE Type = (ENUM_DEAL_TYPE)::HistoryDealGetInteger(Ticket, DEAL_TYPE); \
                                                                                              \
      if ((Type <= DEAL_TYPE_SELL) && (::HistoryDealGetString(Ticket, DEAL_SYMBOL) == Symb))  \
      {                                                                                       \
        const double Price = ::HistoryDealGetDouble(Ticket, DEAL_PRICE);                      \
                                                                                              \
        switch ((ENUM_DEAL_ENTRY)::HistoryDealGetInteger(Ticket, DEAL_ENTRY))                 \
        {                                                                                     \
          case DEAL_ENTRY_OUT:                                                                \
          case DEAL_ENTRY_INOUT:                                                              \
            this.Positions[Amount++] = Type ? Price / this.Positions[Amount]                  \
                                            : this.Positions[Amount] / Price;                 \
                                                                                              \
          case DEAL_ENTRY_IN:                                                                 \
            this.Positions[Amount] = Price;                                                   \
        }                                                                                     \
      }                                                                                       \
    }                                                                                         \
  }                                                                                           \
}

  int GetPositions( const string Symb, const MAGIC_TYPE Magic = -1 )
  {
    int Amount = 0;

  #ifdef __MQL5__
    static const bool IsNetting = ((ENUM_ACCOUNT_MARGIN_MODE)::AccountInfoInteger(ACCOUNT_MARGIN_MODE) != ACCOUNT_MARGIN_MODE_RETAIL_HEDGING);

    #ifdef __VIRTUAL__
      if (VIRTUAL::GetHandle())
        MACROS_MT4
      else
    #endif // __VIRTUAL__
      if (IsNetting)
        MACROS_MT5
      else
      #ifdef __MT4ORDERS__
        MACROS_MT4
      #else // __MT4ORDERS__
        MACROS_MT5
      #endif // __MT4ORDERS__
  #else // __MQL5__
    MACROS_MT4
  #endif // __MQL5__

    return(::ArrayResize(this.Positions, Amount));
  }

#undef MACROS_MT5
#undef MACROS_MT4

#ifdef __SINGLETESTERCACHE__
  int GetPositions( const SINGLETESTERCACHE &SingleTesterCache )
  {
    int Amount = 0;

    const string Symb = SingleTesterCache.Header.symbol[];
    const int Size = ::ArrayResize(this.Positions, ::ArraySize(SingleTesterCache.Deals));

    for (int i = 0; i < Size; i++)
    {
      const TradeDeal Deal = SingleTesterCache.Deals[i];

      if ((Deal.action <= DEAL_TYPE_SELL) &&
          ((Deal.entry == DEAL_ENTRY_OUT) || (Deal.entry == DEAL_ENTRY_INOUT)) &&
          (Deal.symbol[] == Symb))
        this.Positions[Amount++] = Deal.action ? Deal.price_open / Deal.price_close
                                               : Deal.price_close / Deal.price_open;
    }

    return(::ArrayResize(this.Positions, Amount));
  }
#endif // __SINGLETESTERCACHE__

  //        .
  // Markup -   . , 20  : Markup = 20 / 1e6.
  bool GetSumGain( const double Risk,
                   double &SumGain, double &MaxDD, double &RF,
                   const double Markup = 0, const MAGIC_TYPE Magic = -1 ) const
  {
    bool Res = ::ArraySize(this.Positions);
    const double Leverage = Risk * 100;
    const double NewMarkup = (1 - Markup) / (1 + Markup);

    SumGain = 1;
    MaxDD = 0;
    RF = 1;

    double MaxGain = SumGain;
    double DDGain = SumGain;

    const int Size = ::ArraySize(this.Positions);

    for (int i = 0; (i < Size) && (Res = (SumGain > 0)); i++)
    {
      SumGain *= 1 + Leverage * (NewMarkup * this.Positions[i] - 1);

      if (SumGain > MaxGain)
        MaxGain = SumGain;
      else if (1 - SumGain / MaxGain > MaxDD)
      {
        MaxDD = 1 - SumGain / MaxGain;

        DDGain = SumGain;
      }
    }

    RF = SumGain / DDGain;

    return(Res);
  }

  //      .
  double GetRisk( const double NewMaxDD, const double Markup = 0, const double Alpha = 0.01 ) const
  {
    double RiskLeft = 0;
    double RiskRight = 10;

    double Risk = 0;
    double SumGain, MaxDD, RF;

    if (::ArraySize(this.Positions))
      do
      {
        Risk = (RiskLeft + RiskRight) / 2;

        if (!this.GetSumGain(Risk, SumGain, MaxDD, RF, Markup))
        {
          MaxDD = DBL_MAX;

          RiskRight = Risk;
        }
        else if (MaxDD > NewMaxDD)
          RiskRight = Risk;
        else
          RiskLeft = Risk;
      }
      while (!::IsStopped() && (::MathAbs(MaxDD - NewMaxDD) > Alpha));

    return(Risk);
  }

public:
  ONTESTERCUSTOM( void )
  {
  }

  ONTESTERCUSTOM( const string Symb, const MAGIC_TYPE Magic = -1 )
  {
    this.GetPositions(Symb, Magic);
  }

#ifdef __SINGLETESTERCACHE__
  ONTESTERCUSTOM( const SINGLETESTERCACHE &SingleTesterCache )
  {
    this.GetPositions(SingleTesterCache);
  }
#endif // __SINGLETESTERCACHE__

  int Filter( const double Part )
  {
    const int Size = ::ArraySize(this.Positions);
    const int AmountDelete = (int)(Size * Part / 2);

    if ((Part > 0) && (Part < 1) && AmountDelete)
    {
    #define VALUE 0
    #define INDEX 1
      double ArrayIn[][2];

      ::ArrayResize(ArrayIn, Size);

      for (int i = 0; i < Size; i++)
      {
        ArrayIn[i][VALUE] = this.Positions[i];
        ArrayIn[i][INDEX] = i;
      }

      ::ArraySort(ArrayIn);

      double ArrayOut[][2];

      ::ArrayResize(ArrayOut, Size);

      for (int i = 0; i < Size; i++)
      {
        ArrayOut[i][VALUE] = ArrayIn[i][INDEX];
        ArrayOut[i][INDEX] = ((i < AmountDelete) || (i >= Size - AmountDelete)) ? 0 : ArrayIn[i][VALUE];
      }

      ::ArraySort(ArrayOut);

      int Amount = 0;

      for (int i = 0; i < Size; i++)
        if (ArrayOut[i][INDEX])
          this.Positions[Amount++] = ArrayOut[i][INDEX];

      ::ArrayResize(this.Positions, Amount);

    #undef INDEX
    #undef VALUE
    }

    return(::ArraySize(this.Positions));
  }

  double TesterStatistics( const ENUM_ONTESTERCUSTOM Property, const double Value, const double Markup = 0 ) const
  {
    double Res = 0;

    double SumGain;
    double RF;

    double MaxDD = Value;
    double Risk = Value;

  #define MACROS(A)                                                        \
    Res = this.GetSumGain(Risk, SumGain, MaxDD, RF, Markup / 1e6) ? A : 0; \
                                                                           \
    break;

    switch (Property)
    {
      case ONTESTERCUSTOM_GAIN:                   // Gain
        MACROS(SumGain)
      case ONTESTERCUSTOM_MAXDD:                  // MaxDD
        MACROS(MaxDD)
      case ONTESTERCUSTOM_RECOVERYFACTOR:         // Recovery Factor
        MACROS(RF)
      case ONTESTERCUSTOM_GAIN_FIXDD:             // Gain with Fix DD
        Risk = this.GetRisk(MaxDD, Markup / 1e6);
        MACROS(SumGain)
      case ONTESTERCUSTOM_RISK_FIXDD:             // Risk with Fix DD
        Res = this.GetRisk(MaxDD, Markup / 1e6);

        break;
      case ONTESTERCUSTOM_RECOVERYFACTOR_FIXDD:   // Recovery Factor with Fix DD
        Risk = this.GetRisk(MaxDD, Markup / 1e6);
        MACROS(RF)
    }

  #undef MACROS

    return(Res);
  }

  void operator =( const ONTESTERCUSTOM &OnTesterCustom )
  {
    ::ArrayFree(this.Positions);

    ::ArrayCopy(this.Positions, OnTesterCustom.Positions);

    return;
  }
};